#include <os9/os9.h>

#include <assert.h>
#include <errno.h>
#include <string.h>

#include <libk.h>

#include "sysreg.h"
#include "arch.h"

/* TCR_EL1 */
#define NS 1ull
#define OS 2ull
#define IS 3ull
#define NC 0ull
#define WBA 1ull
#define WTA 2ull
#define WT 3ull
#define G16K 1ull
#define G4K 2ull
#define G64K 3ull

#define I4GB 0ull
#define I64GB 1ull
#define I1TB 2ull
#define I4TB 3ull
#define I16TB 4ull
#define I256TB 5ull

#define TBI1 (1ull << 38)
#define TBI0 (1ull << 37)
#define AS (1ull << 36)
#define IPS(x) ((x) << 32)
#define TG1(x) (((x)&3) << 30)
#define SH1(x) ((x) << 28)
#define ORGN1(x) ((x) << 26)
#define IRGN1(x) ((x) << 24)
#define EDP1 (1ull << 23)
#define A1 (1ull << 22)
#define T1SZ(x) (((x)&63) << 16)
#define TG0(x) (((x)&3) << 14)
#define SH0(x) ((x) << 12)
#define ORGN0(x) ((x) << 10)
#define IRGN0(x) ((x) << 8)
#define EDP0 (1ull << 7)
#define T0SZ(x) (((x)&63) << 0)

/* MMU */
#define RDWR  0
#define EL0 1
#define RD 2


/* table attributes */
#define APTable(x) ((pte_t) (x) << 62)
#define UXNTable(x) ((pte_t) (x) << 60)
#define PXNTable(x) ((pte_t) (x) << 59)

/* page high attributes */
#define UXN ((pte_t) 1 << 54)
#define PXN ((pte_t) 1 << 53)
#define CONT ((pte_t) 1 << 52)

/* page low attributes */
/*
 * TODO: Define macros for OS(), because we are using
 * the same macros than in OSx() but Non shareable has
 * a different representation
 */
#define NG(x) ((pte_t) (x) << 11)
#define AF(x) ((pte_t) (x) << 10)
#define SH(x) ((pte_t) (x) << 8)
#define AP(x) ((pte_t) (x) << 6)
#define IDX(x) ((pte_t) (x) << 2)

#define VASIZE  40
#define MASK(pa) ((pa) & (-1ull>>(64-VASIZE) & ~0ull<<12))
#define INVALID 0
#define BLOCK 1
#define TABLE 3
#define PAGE 3

#define TINDEX(lvl, va) ((va)>>(39-(lvl)*9) & 0x1FF)

#define DEV 7
#define MEM 0

/* ID_AA64MMFR0_EL1 */
#define TGRAN4(x) ((x)>>28 & 0xf)
#define TGRAN64(x) ((x)>>24 & 0xf)
#define TGRAN16(x) ((x>>20 & 0xf)
#define PAR(x) ((x) & 0xf)

/* SPSR */
#define NF(x) ((unsigned long) (x) << 31)
#define ZF(x) ((unsigned long) (x) << 30)
#define CF(x) ((unsigned long) (x) << 29)
#define VF(x) ((unsigned long) (x) << 28)
#define SS(x) ((unsigned long) (x) << 21)
#define IL(x) ((unsigned long) (x) << 20)

#define DAIF(x) ((unsigned long) (x) << 6)
#define MODE(x) ((unsigned long) (x) << 0)

#define A64 0
#define A32 1

#define AARCH(x) ((unsigned long) (x) << 4)

#define EL0t 0
#define EL1t 4
#define EL1h 5

/* MAIR */
#define DEVICE(x) (x)
#define NORMAL(c, t, a) ((c) | (t) | (a))

#define nGnRnE 0ull
#define nGnRE 2ull
#define nGRE 4ull
#define GRE 8ull

/*
 * TODO: Improve these macros
 */
#define ONC 4ull
#define OWT 0ull
#define OWB 4ull

#define NTRANSIENT 8ull
#define RALLOC 2ull
#define WALLOC 1ull

#define TRANSIENT 0ull
#define NRALLOC 0ull
#define NWALLOC 0ull

int inlowmem = 1;

static pte_t
newpage(void)
{
	void *bp;

	bp = allocb(1);
	memset(bp, 0, PAGESIZE);

	return (pte_t) bp;
}

static pte_t *
walker(uintptr_t va)
{
	pte_t *bp, e;
	int lvl;
	pte_t ttbr;

	ttbr = (va & 1ull<<63) ? sysrd(TTBR1_EL1) : sysrd(TTBR0_EL1);
	bp = (pte_t *) MASK(ttbr);
	va = MASK(va);

	for (lvl = 0; lvl < 3; lvl++) {
		int n = TINDEX(lvl, va);

		e = bp[n];
		if (e == INVALID) {
			e = newpage();
			bp[n] = e | TABLE;
			barrier(DATA);
		}
		bp = (pte_t *) MASK(e);
	}

	return &bp[TINDEX(3, va)];
}

void
vmap(phyaddr_t pa, uintptr_t va, int perm)
{
	static mutex_t m;
	pte_t *p, attr;

	attr = AF(1) | SH(OS) | PAGE;
	attr |= (perm & MW) ? AP(RDWR) : AP(RD);
	attr |= (perm & MD) ? IDX(DEV) : IDX(MEM);

	if ((perm & MX) == 0)
		attr |= UXN|PXN;

	if (va & 1ull<<63)
		attr |= EL0;

	/*
	 * TODO: It does not need locking, it needs a exclusive load
	 * locking is needed if we are in a system
	 * with SMMUv3 and page tables are shared.
	 */
	lock(&m);
	p = walker(va);

	/*
	 * Here is needed to do a break-before-make as
	 * the architecture specifies to avoid a race condition
	 * in multi processor systems.
	 */
	*p = INVALID;
	invtlb(va);
	barrier(DATA);
	barrier(CODE);

	*p = MASK(pa) | attr;
	barrier(DATA);
	unlock(&m);

	dbg("page %p: %llx\n", p, MASK(pa) | attr);
}

void
mapseg(Map *mp)
{
	phyaddr_t pa;
	uintptr_t va;
	unsigned long i, n;

	assert(mp->perm & MR);

	dbg("map %s: %llx -> %llx %c %c%c%c (%llu)\n",
		(mp->name) ? mp->name : "unnamed",
		mp->va, mp->pa,
		(mp->perm & MD) ? 'd' : 'n',
		(mp->perm & MR) ? 'r' : '-',
		(mp->perm & MW) ? 'w' : '-',
		(mp->perm & MX) ? 'x' : '-',
		mp->siz);

	pa = mp->pa;	
	va = mp->va;

	n =  (mp->siz + 4*KiB-1)/(4*KiB);

	for (i = 0; i < n; i++) {
		vmap(pa, va, mp->perm);
		pa += 4*KiB;
		va += 4*KiB;
	}
}

void
taskmap(void)
{
	uint64_t tcr;

	tcr = sysrd(TCR_EL1);
	tcr &= ~T0SZ(-1);
	tcr |= T0SZ(24);
	syswr(TCR_EL1, tcr);
	barrier(CODE);

	/*
	 * TODO: setup SCTLR_EL1.WXN
	 */
}

void
kernelmap(void)
{
	Map *mp;
	uint64_t tcr;

	tcr = sysrd(TCR_EL1);
	tcr &= ~T1SZ(-1);
	tcr |= T1SZ(24);
	syswr(TCR_EL1, tcr);
	barrier(CODE);

	/*
	 * FIXME: We are using a full page when
	 * we only need 16 bytes with 16 byte alignment.
	 */
	syswr(TTBR1_EL1, newpage());

	dbg("TTBR1_EL1=%llx\n", sysrd(TTBR1_EL1));

	for (mp = mach.maps; mp->siz != 0; mp++)
		mapseg(mp);
	barrier(DATA);
	barrier(CODE);
}

void
initmap(void)
{
	pte_t *bp, tbl;
	phyaddr_t pa;
	uint64_t tcr;
	pte_t attr = CONT | AF(1) | SH(OS) | AP(RDWR) | IDX(DEV) | BLOCK;

	tcr = sysrd(TCR_EL1);
	tcr &= ~T0SZ(-1);
	tcr |= T0SZ(32);
	syswr(TCR_EL1, tcr);
	barrier(CODE);

	tbl = newpage();
	bp = (pte_t *) tbl;
	for (pa = 0; pa < 4*GiB; pa += 1*GiB)
		*bp++ = pa | attr;

	syswr(TTBR0_EL1, tbl);
	barrier(DATA);
	barrier(CODE);
}

void
mmuon(void)
{
	uint64_t sctlr;

	sctlr = sysrd(SCTLR_EL1);
	sctlr |= M;
	syswr(SCTLR_EL1, sctlr);
	barrier(CODE);
}

void
immu(void)
{
	uint64_t tcr, mair, id;

	id = sysrd(ID_AA64MMFR0_EL1);
	if (TGRAN4(id) != 0 || PAR(id) < I1TB)
		panic("unsupported mmu");

	/* FIXME: Add Inner properties */
	mair =
		DEVICE(nGnRnE) << 56 |
		DEVICE(nGnRE) << 48 |
		DEVICE(nGRE) << 40 |
		DEVICE(GRE) << 32 |
		NORMAL(OWT, TRANSIENT, RALLOC|WALLOC) << 24 |
		NORMAL(OWB, TRANSIENT, RALLOC|WALLOC) << 16 |
		NORMAL(OWT, NTRANSIENT, RALLOC|WALLOC) << 8 |
		NORMAL(OWB, NTRANSIENT, RALLOC|WALLOC) << 0;

	tcr =
		IPS(I1TB) |
		TG1(G4K) | TG0(G4K) |
		SH1(OS) | SH0(OS) |
		ORGN1(WBA) | ORGN0(WBA) |
		IRGN1(WBA) | IRGN0(WBA);

	invalltlb();
	syswr(MAIR_EL1, mair);
	syswr(TCR_EL1,  tcr);
	barrier(CODE);
}
