#include <os9/os9.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "dev.h"

#define NR_ARS     2

#define ARMAG      "!<arch>\n"     /* ar "magic number" */
#define SARMAG     8               /* strlen(ARMAG); */
#define ARFMAG     "`\n"
#define SARNAM     16


struct ar_hdr {
	char ar_name[SARNAM];      /* name */
	char ar_date[12];          /* modification time */
	char ar_uid[6];            /* user id */
	char ar_gid[6];            /* group id */
	char ar_mode[8];           /* octal file permissions */
	char ar_size[10];          /* size in bytes */
	char ar_fmag[2];           /* consistency check */
};

struct arfile {
	Chan *c;
	long offset[NR_FILES];
	long size[NR_FILES];
};

static struct arfile archives[NR_ARS];
static int nars;


static int
gethdr(Chan *c, struct ar_hdr *hdr)
{
	int n, ch, i;

	n = devtab[c->type]->read(c, hdr, sizeof(*hdr));
	if (n <= 0)
		return n;

	if (n != sizeof(*hdr) || strncmp(hdr->ar_fmag, ARFMAG, 2) != 0) {
		errno = EINVAL;
		return -1;
	}

	for (i = SARNAM-1; i >= 0; i--) {
		ch = hdr->ar_name[i];
		if (ch != ' ' && ch != '/')
			break;
		hdr->ar_name[i] = '\0';
	}

	return 1;
}

static int
argen(Chan *c, Dirtab *tab, int ntab, int n, Dir *dir)
{
	int i, r, len;
	long off;
	Chan nc;
	struct ar_hdr hdr;
	struct arfile *ar;

	if (c->dev >= nars)
		panic("argen");
	clone(archives[c->dev].c, &nc);
	ar = &archives[nc.dev];

	for (i = 0; i <= n; i++) {
		if (ar->offset[i] == -1)
			return -1;
		off = ar->offset[i];
		if (devtab[c->type]->seek(&nc, off, SEEK_SET) < 0)
			return -1;
		if ((r = gethdr(&nc, &hdr)) <= 0)
			return r;
	}

	if ((len = strlen(hdr.ar_name)) >= NAMELEN) {
		errno = ENAMETOOLONG;
		return -1;
	}

	mkentry(c, dir,
	        hdr.ar_name, atol(hdr.ar_size),
	        QID(CHFILE, 0, n),
	        O_READ);

	return 1;
}

static int
arwalk(Chan *c, char *name)
{
	return devwalk(c, name, NULL, 0, argen);
}

static int
arstat(Chan *c, char *file, unsigned char *buf, int n)
{
        return devstat(c, file, buf, n, NULL, 0, argen);
}

static int
arread(Chan *c, void *buf, int n)
{
	long off;
	Chan cs;
	struct arfile *ar;
	long size;
	long long path;

	if (c->qid.type == CHDIR)
		return dirread(c, buf, n, NULL, 0, argen);

	if (c->dev >= nars)
		panic("arread1");
	ar = &archives[c->dev];

	path = c->qid.path;
	if (path >= NR_FILES || ar->offset[path] == -1)
		panic("arread2");
	clone(ar->c, &cs);

	size = ar->size[path];
	if (c->offset >= size)
		return 0;

	if (n < 0) {
		errno = EINVAL;
		return -1;
	}
	if (n > size - c->offset)
		n = size - c->offset;

	off = ar->offset[path] + c->offset + sizeof(struct ar_hdr);
	if (devtab[cs.type]->seek(&cs, off, SEEK_SET) < 0)
		return -1;

	n = devtab[cs.type]->read(&cs, buf, n);
	if (n > 0)
		c->offset += n;

	return n;
}

static Chan *
armount(Chan *c, char *spec)
{
	int r, n, t;
	long len, offhdr;
	Chan *cspec;
	char buf[SARMAG];
	struct ar_hdr hdr;
	struct arfile *ar;

	if (nars == NR_ARS) {
		errno = ENOMEM;
		return NULL;
	}
	ar = &archives[nars];

	for (n = 0; n < NR_FILES; n++)
		ar->offset[n] = -1;

	if ((cspec = namec(spec, O_READ)) == NULL)
		return NULL;
	ar->c = cspec;

	r = devtab[cspec->type]->read(cspec, buf, SARMAG);
	if (r < 0)
		goto err;
	if (r != SARMAG || strncmp(buf, ARMAG, SARMAG)) {
		errno = EINVAL;
		goto err;
	}

	for (n = 0; n < NR_FILES; n++) {
		offhdr = cspec->offset;
		switch (gethdr(cspec, &hdr)) {
		case 0:
			return attach('R', nars++);
		case -1:
			goto err;
		default:
			ar->offset[n] = offhdr;
			len = atol(hdr.ar_size);
			ar->size[n] = len;
			if ((len % 2) != 0)
				len++;

			t = cspec->type;
			if (devtab[t]->seek(cspec, len, SEEK_CUR) < 0)
				goto err;
			break;
		}
	}
	errno = ENOMEM;

err:
	delchan(cspec);
	return NULL;
}

Dev ardevtab = {
	.id = 'R',
	.stat = arstat,
	.clone = devclone,
	.attach = devattach,
	.walk = arwalk,
	.read = arread,
	.write = deverrwrite,
	.mount = armount,
	.seek = devseek,
	.sync = devsync,
};
