#include <os9/os9.h>

#include <libk.h>

#include <ctype.h>
#include <errno.h>
#include <string.h>

#include "dev.h"

#define CONSOUT      2
#define NAMESIZE     15
#define CONSSTATUS   128

#define CTRL(x) ((x) & 0x1f)
#define DEL 127

enum Orootqid {
	Qconsfs,
	Qraw,
	Qctl
};

struct conscmd {
	char *name;
	int (*fn)(struct conscmd *, char *);
};

static struct conscmd cmds[];

static Dirtab dirtab[] = {
	{"raw", QID(CHFILE, 0, Qraw), 0, O_READ | O_WRITE},
	{"ctl", QID(CHFILE, 0, Qctl), 0, O_READ | O_WRITE}
};

static Chan *in;
static char inname[NAMESIZE];
static Chan *out[CONSOUT];
static char outname[CONSOUT][NAMESIZE];

static char buffer[LINELEN];
static int head;
static int tail;
static int cookedf, echof;

static int
conswalk(Chan *c, char *name)
{
	return devwalk(c, name, dirtab, NELEM(dirtab), devgen);
}

static int
consstat(Chan *c, char *file, unsigned char *buf, int n)
{
	return devstat(c, file, buf, n, dirtab, NELEM(dirtab), devgen);
}

static int
consstatus(void *buf, Chan *c, int n)
{
	int i;
	int len = 0;
	char tmp[CONSSTATUS];

	if (in)
		len += ksnprint(tmp, sizeof(tmp),
				"addin %s\n",
				inname);

	for (i = 0; i < CONSOUT; i++) {
		if (out[i]) {
			len += ksnprint(tmp + len,
			                sizeof(tmp) - len,
					"addout %s\n",
					outname[i]);
		}
	}

	if (len == sizeof(tmp))
		panic("consstatus");

	return buf2chan(c, buf, tmp, n, len);
}

static int
flushraw(char *buf, int n)
{
	Chan **pp, *p;
	int r, w;

	r = 0;
	for (pp = out; pp < &out[CONSOUT]; pp++) {
		if ((p = *pp) == NULL)
			continue;
		w = devtab[p->type]->write(p, buf, n);
		if (w < 0) {
			r = -1;
		} else if (w != n) {
			errno = EIO;
			r = -1;
		}
	}

	return (r < 0) ? r : n;
}

static int
conswriteraw(char *buf, int n)
{
	int i, idx;
	char wrtbuf[LINELEN];

	idx = 0;
	for (i = 0; i < n; i++) {
		if (idx + 2 > sizeof(wrtbuf)) {
			if (flushraw(wrtbuf, idx) < 0)
				return -1;
			idx = 0;
		}

		if (buf[i] == '\n')
			wrtbuf[idx++] = '\r';

		wrtbuf[idx++] = buf[i];
	}

	if (idx > 0) {
		if (flushraw(wrtbuf, idx) < 0)
			return -1;
	}
	return n;
}

static int
consaddin(struct conscmd *cmd, char *s)
{
	if (in) {
		errno = ENOMEM;
		return -1;
	}

	in = namec(s, O_READ);
	strcpy(inname, s);
	return 0;
}

static int
consaddout(struct conscmd *cmd, char *s)
{
	int i;

	for (i = 0; i < CONSOUT; i++) {
		if (!out[i]) {
			out[i] = namec(s, O_WRITE);
			strcpy(outname[i], s);
			return 0;
		}
	}
	errno = ENOMEM;
	return -1;
}

static int
consdelin(struct conscmd *cmd, char *s)
{
	if (!in || strcmp(s, inname)) {
		errno = ENOENT;
		return -1;
	}

	delchan(in);
	in = NULL;
	inname[0] = '\0';
	return 0;
}

static int
consdelout(struct conscmd *cmd, char *s)
{
	int i;

	for (i = 0; i < CONSOUT; i++) {
		if (out[i] && !strcmp(s, outname[i])) {
			delchan(out[i]);
			out[i] = NULL;
			outname[i][0] = '\0';
			return 0;
		}
	}
	errno = ENOENT;
	return -1;
}

static int
consmode(struct conscmd *cmd, char *s)
{
	int mode;

	if (!strcmp(s, "on")) {
		mode = 1;
	} else if(!strcmp(s, "off")) {
		mode = 0;
	} else {
		errno = EINVAL;
		return -1;
	}

	if (!strcmp(cmd->name, "cooked")) {
		cookedf = mode;
	} else if (!strcmp(cmd->name, "echo")) {
		echof = mode;
	} else {
		errno = EINVAL;
		return -1;
	}

	return 0;
}

static int
conswrite(Chan *c, void *buf, int n)
{
	struct conscmd *cmd;
	char *tokens[2];

	switch (c->qid.path) {
	case Qraw:
		return conswriteraw(buf, n);
	case Qctl:
		if (tokenize(buf, n, tokens, 2) != 2) {
			errno = EINVAL;
			return -1;
		}

		for (cmd = cmds; *cmd->name; ++cmd) {
			if (!strcmp(tokens[0], cmd->name))
				return (*cmd->fn)(cmd, tokens[1]);
		}

		errno = EINVAL;
		return -1;
	default:
		panic("conswrite");
	}
}

static int
echo(char *buf, int n)
{
	if (echof) {
		if (conswriteraw(buf, n) < 0)
			return -1;
	}
	return buf[n-1];
}

static void
erase(int n)
{
	int i;
	char ch;

	if (head < n)
		n = head;

	ch = '\b';
	for (i = 0; i < n; i++)
		echo(&ch, 1);
	ch = ' ';
	for (i = 0; i < n; i++)
		echo(&ch, 1);
	ch = '\b';
	for (i = 0; i < n; i++)
		echo(&ch, 1);

	head -= n;
}

static int
wordsize(void)
{
	int n = 1;

	while (head - n > 0 && buffer[head - n] == ' ')
		n++;
	while (head - n > 0 && buffer[head - n] != ' ')
		n++;

	return n;
}

static int
edit(char key)
{
	char ch;

	if (!cookedf) {
		buffer[head++] = key;
		return echo(&key, 1);
	}

	switch (key) {
	case '\0':
		return 0;
	case '\b':
	case DEL:
		erase(1);
		return 0;
	case CTRL('L'):
		ch = '\r';
		echo(&ch, 1);
		echo(buffer, head-1);
		return 0;
	case CTRL('U'):
		ch = '\r';
		echo(&ch, 1);
		erase(head);
		echo(&ch, 1);
		return 0;
	case CTRL('W'):
		erase(wordsize());
		return 0;
	case '\r':
		key = '\n';
	default:
		buffer[head++] = key;
		return echo(&key, 1);
	}
}

static int
fillbuffer(void)
{
	char ch;

	while (head + 1 < sizeof(buffer)) {
		if (devtab[in->type]->read(in, &ch, 1) < 0) {
			ch = EOF;
			break;
		}
		if (edit(ch) == '\n')
			break;
	}
	if (ch == EOF && head == 0) {
		errno = EIO;
		return -1;
	}

	return 0;
}

static int
readqueue(void *buf, int n)
{
	int size = head - tail;

	if (n < size) {
		memcpy(buf, &buffer[tail], n);
		tail += n;
		return n;
	} else {
		memcpy(buf, &buffer[tail], size);
		head = 0;
		tail = 0;
		return size;
	}
}

static int
consread(Chan *c, void *buf, int n)
{
	switch (c->qid.path) {
	case Qconsfs:
		return dirread(c, buf, n, dirtab, NELEM(dirtab), devgen);
	case Qraw:
		if (!in) {
			errno = ENOENT;
			return -1;
		}
		if (head == 0) {
			if (fillbuffer() < 0)
				return -1;
		}

		return readqueue(buf, n);
	case Qctl:
		return consstatus(buf, c, n);
	default:
		panic("consread");
	}
}

static struct conscmd cmds[] = {
	{"addin", consaddin},
	{"addout", consaddout},
	{"delin", consdelin},
	{"delout", consdelout},
	{"cooked", consmode},
	{"echo", consmode},
	{""}
};

Dev consdevtab = {
	.id = 'c',
	.stat = consstat,
	.clone = devclone,
	.attach = devattach,
	.walk = conswalk,
	.read = consread,
	.write = conswrite,
	.seek = deverrseek,
	.sync = devsync,
};
