#include <os9/os9.h>

#include <ctype.h>
#include <errno.h>
#include <libk.h>
#include <stdlib.h>
#include <string.h>

#include "dev.h"
#include "uart.h"

#define UARTSTATUS   128
#define NR_UARTS     2

enum Ouartqid {
	Quartfs,
	Qraw,
	Qctl,
};

static Dirtab dirtab[] = {
	{"raw", QID(CHFILE, 0, Qraw), 0, O_READ | O_WRITE},
	{"ctl", QID(CHFILE, 0, Qctl), 0, O_READ | O_WRITE}
};

static Uart *uarts[NR_UARTS];
static int nuarts;

static int
uartwalk(Chan *c, char *name)
{
	return devwalk(c, name, dirtab, NELEM(dirtab), devgen);
}

static int
uartstat(Chan *c, char *file, unsigned char *buf, int n)
{
	return devstat(c, file, buf, n, dirtab, NELEM(dirtab), devgen);
}

static Uart *
getuart(Chan *c)
{
	if (c->dev >= NR_UARTS)
		panic("getuart");
	return uarts[c->dev];
}

static int
uartstatus(Uart *up, void *buf, Chan *c, int n)
{
	int len;
	struct uartstat st;
	char tmp[UARTSTATUS];

	(*up->phy->status)(up, &st);

	len = ksnprint(tmp, sizeof(tmp),
	               "b%ld c%d d%d e%d l%d m%d p%c r%d s%d i%d\n",
	               up->rate,
	               (up->hang & SIG_DCD) != 0,
	               up->dtr,
	               (up->hang & SIG_DSR) != 0,
	               up->nbits,
	               up->ctsflow,
	               up->parity,
	               up->rts,
	               up->nstop,
	               up->fifo);

	if (len == sizeof(tmp))
		panic("uartstatus");

	return buf2chan(c, buf, tmp, n, len);
}

static void
uartdrain(Uart *up)
{
	/* TODO */
}

/*
 * bn   Set the baud rate to n.
 * cn   Set hangup on DCD if n is non-zero; else clear it.
 * dn   Set DTR if n is non-zero; else clear it.
 * f    Flush output queue.
 * in   Enable/disable the FIFOs.  If n is zero the FIFOs are
 *      disabled; otherwise n is taken as a trigger level for
 *      the FIFOs.  The trigger levels supported are device
 *      dependant, but usually include 1, 4 and 8.  An unrecog-
 *      nised, but non-zero, value of n causes the maximum-
 *      supported trigger level to be set.
 * en   Set hangup on DSR if n is non-zero; else clear it.
 * ln   Set number of bits per byte to n. Legal values are 5,
 *      6, 7, or 8.
 * mn   Obey modem CTS signal if n is non-zero; else clear it.
 * pc   Set parity to odd if c is o, to even if c is e; else
 *      set no parity.
 * rn   Set RTS if n is non-zero; else clear it.
 * sn   Set number of stop bits to n. Legal values are 1 or 2.
*/
static int
uartctl(Uart *up, char *cfg, int n)
{
	int cnt, r, i;
	Uartphy *phy = up->phy;
	unsigned char *s = (unsigned char *) cfg;

	for (cnt = i = 0; i < n; cnt = ++i) {
		int sig, par, c = s[i];

		if (isspace(c))
			continue;

		par = 0;
		while (i+1 < n && isdigit(s[i+1]))
			par = par*10 + s[++i] - '0';

		switch (c) {
		case 'b':
			uartdrain(up);	
			r = (*phy->setbauds)(up, par);
			break;
		case 'e':
			up->hang &= ~ (1 << SIG_DSR);
			up->hang |= (par != 0) << SIG_DSR;
			r = 0;
			break;
		case 'c':
			up->hang &= ~ (1 << SIG_DCD);
			up->hang |= (par != 0) << SIG_DCD;
			r = 0;
			break;
		case 'r':
			par = (par != 0) << SIG_RTS;
			sig = SIG_RTS;
			goto signal;
		case 'd':
			par = (par != 0) << SIG_DCD;
			sig = SIG_DCD;
		signal:
			r = (*phy->signal)(up, sig, par);
			break;
		case 'f':
			 r = (*phy->flush)(up);
			break;
		case 'i':
			r = (*phy->setfifo)(up, par);
			break;
		case 'l':
			 r = (*phy->numbits)(up, par);
			break;
		case 'm':
			r = (*phy->obey)(up, SIG_CTS, par != 0);
			break;
		case 'p':
			if (par != OPARITY && par != EPARITY)
				par = NPARITY;
			r = (*phy->parity)(up, par);
			break;
		case 's':
			if (par != 1 && par != 2)
				return -1;
			r = (*phy->bitstop)(up, par);
			break;
		default:
			r = -1;
		}

		if (r < 0)
			return cnt;
	}

	return cnt;
}

static int
uartread(Chan *c, void *buf, int n)
{
	Uart *up;

	switch (c->qid.path) {
	case Quartfs:
		return dirread(c, buf, n, dirtab, NELEM(dirtab), devgen);
	case Qraw:
		up = getuart(c);
		return (*up->phy->read)(up, buf, n);
	case Qctl:
		up = getuart(c);
		return uartstatus(up, buf, c, n);
	default:
		panic("uartread");
	}
}

static int
uartwrite(Chan *c, void *buf, int n)
{
	Uart *up;

	switch (c->qid.path) {
	case Qraw:
		up = getuart(c);
		return (*up->phy->write)(up, buf, n);
	case Qctl:
		up = getuart(c);
		return uartctl(up, buf, n);
	default:
		panic("uartwrite");
	}
}

void
uartlink(Uartphy *phy, Attr *attr)
{
	size_t siz;
	Uart *up;
	Attr *a;
	char *cfg = NULL;

	if (nuarts == NR_UARTS)
		panic("uartlink1");

	siz = sizeof(Uart);
	up = alloc(siz);
	up->phy = phy;
	uarts[nuarts++] = up;

	for (a = attr; a->key; a++) {
		if (!strcmp(a->key, "base"))
			up->base = (void *) strtoull(a->value, NULL, 0);
		else if (!strcmp(a->key, "clk"))
			up->clk = strtoull(a->value, NULL, 0);
		else if (!strcmp(a->key, "cfg"))
			cfg = a->value;
	}
	(*phy->init)(up, attr);

	if (!cfg)
		cfg = "b19200 l8";
	siz = strlen(cfg);

	if (uartctl(up, cfg, siz) != siz)
		panic("uartlink2");
}

Dev uartdevtab = {
	.id = 't',
	.stat = uartstat,
	.clone = devclone,
	.attach = devattach,
	.walk = uartwalk,
	.read = uartread,
	.write = uartwrite,
	.mount = deverrmount,
	.seek = deverrseek,
	.sync = devsync,
};
