#include <os9/os9.h>

#include "dev.h"
#include "uart.h"

enum phyregs {
	UARTDR,      /* Data register */
	UARTRSR,     /* Receive status register */
	RES1,        /* Reserved */
	RES2,        /* Reserved */
	RES3,        /* Reserved */
	RES4,        /* Reserved */
	UARTFR,      /* Flag register */
	RES5,        /* Reserved */
	UARTILPR,    /* IrDA low-power count register */
	UARTIBRD,    /* Integer baud rate */
	UARTFBRD,    /* Fractional baud rate register */
	UARTLCR_H,   /* Line control register */
	UARTCR,      /* Control register */
	UARTIFLS,    /* Interrupt FIFO level select register */
	UARTIMSC,    /* Interrupt mask set/clear register */
	UARTRIS,     /* Raw interrupt status register */
	UARTMIS,     /* Masked interrupt register */
	UARTICR,     /* Interrupt clear register */
	UARTDMACR,   /* DMA control register */
	RES6,        /* Reserved */
	RES7,        /* Reserved */
	RES8,        /* Reserved */
	RES9,        /* Reserved */
	UARTPID0,    /* UART Periph ID0 */
	UARTPID1,    /* UART Periph ID1 */
	UARTPID2,    /* UART Periph ID2 */
	UARTPID3,    /* UART Periph ID3 */
	UARTCID0,    /* UART Cell ID0 */
	UARTCID1,    /* UART Cell ID1 */
	UARTCID2,    /* UART Cell ID2 */
	UARTCID3,    /* UART Cell ID3 */
};

enum eoi {
	RIMC   = 1 << 0,    /* clears UARTRIINTR */
	CTSMIC = 1 << 1,    /* clears UARTCTSINTR */
	DCDMIC = 1 << 2,    /* clears UARTDCDINTR */
	DSRMIC = 1 << 3,    /* clears UARTDSRINTR */
	RXIC   = 1 << 4,    /* clears UARTRXINTR */
	TXIC   = 1 << 5,    /* clears UARTTXINTR */
	RTIC   = 1 << 6,    /* clears UARTRTINTR */
	FEIC   = 1 << 7,    /* clears UARTFEINTR */
	PEIC   = 1 << 8,    /* clears UARTPEINTR */
	BEIC   = 1 << 9,    /* clears UARTBEINTR */
	OEIC   = 1 << 10,   /* clears UARTOEINTR */
};

#define ALLIC (RIMC|CTSMIC|DCDMIC|DSRMIC|RXIC|TXIC|\
               RTIC|FEIC|PEIC|BEIC|OEIC)

enum uartlcr_h {
	BRK    = 1 << 0,    /* send break */
	PEN    = 1 << 1,    /* Parity enable */
	EPS    = 1 << 2,    /* Even parity enable */
	STP2   = 1 << 3,    /* Two stop bits */
	FEN    = 1 << 4,    /* Enable FIFOs */
	WLEN_5 = 0 << 5,    /* word lenght (5 bits) */
	WLEN_6 = 1 << 5,    /* word lenght (6 bits) */
	WLEN_7 = 2 << 5,    /* word lenght (7 bits) */
	WLEN_8 = 3 << 5,    /* word lenght (8 bits) */
	WLEN   = 3 << 5,    /* clean mask for word length */
	SPS    = 1 << 7,    /* Stick parity select */
};

enum uartcr {
	UARTEN = 1 << 0,    /* UART enable */
	SIREN  = 1 << 1,    /* SIR enable */
	SIRLP  = 1 << 2,    /* SIR low power IrDA mode */
	LBE    = 1 << 7,    /* Loopback enable */
	TXE    = 1 << 8,    /* TX enable */
	RXE    = 1 << 9,    /* RX enable */
	DTR    = 1 << 10,   /* Data transmit ready */
	RTS    = 1 << 11,   /* Request to send */
	OUT1   = 1 << 12,   /* can be used for DCD */
	OUT2   = 1 << 13,   /* can be used for RI */
	RSTEN  = 1 << 14,   /* RTS hardware flow control mode */
	CTSEN  = 1 << 15,   /* CTS hardware flow control mode */
};

#define ALLCR (UARTEN|SIREN|SIRLP|LBE|TXE|RXE|DTR|\
               RTS|OUT1|OUT2|RSTEN|CTSEN)

enum uartfr {
	FR_CTS  = 1 << 0,     /* CTS flag */
	FR_DSR  = 1 << 1,     /* DSR flag */
	FR_DCD  = 1 << 2,     /* DCD flag */
	FR_BUSY = 1 << 3,     /* UART busy transmiting */
	FR_RXFE = 1 << 4,     /* Receive FIFO empty */
	FR_TXFF = 1 << 5,     /* Transmit FIFO full */
	FR_RXFF = 1 << 6,     /* Receive FIFO full */
	FR_TXFE = 1 << 7,     /* Transmit FIFO empty */
	FR_RI   = 1 << 8,     /* Ring indicator */
};

enum uartifls {
	TXIFLSEL = 0,
	RXIFLSEL = 3,
};

static unsigned long
plin(Uart *up, int reg)
{
	char *addr = up->base;

	addr += reg * 4;
	return inm32(addr);
}

static void
plout(Uart *up, int reg, unsigned long val)
{
	char *addr = up->base;

	addr += reg * 4;
	outm32(val, addr);
}

static int
setbauds(Uart *up, long rate)
{
	unsigned div;
	unsigned long cr;

	div = (up->clk * 4) / rate;
	if (div == 0)
		return -1;

	cr = plin(up, UARTCR);
	plout(up, UARTCR, cr & ~UARTEN);
	plout(up, UARTIBRD, div >> 6);
	plout(up, UARTFBRD, div & 0x3f);
	up->rate = rate;
	plout(up, UARTCR, cr | UARTEN);

	return 0;
}

static void
pl011init(Uart *up, Attr *attr)
{
	unsigned long cr;

	cr = plin(up, UARTCR) & ~ALLCR;
	plout(up, UARTCR, cr);
	plout(up, UARTICR, ALLIC);
	plout(up, UARTRSR, 0);
	plout(up, UARTLCR_H, WLEN_8);
	plout(up, UARTIMSC, 0);
	plout(up, UARTIBRD, 0);
	plout(up, UARTFBRD, 0);
	plout(up, UARTCR, cr | UARTEN | TXE | RXE);

	up->rate = 0;
	up->nbits = 8;
	up->nstop = 1;
	up->hang = 0;
	up->ctsflow = 0;
	up->parity = NPARITY;
	up->fifo = 0;
}

static int
pl011read(Uart *up, void *buf, int n)
{
	int i;
	char *bp = buf;

	for (i = 0; i < n; i++) {
		while ((plin(up, UARTFR) & FR_RXFE) != 0)
			;
		bp[i] = plin(up, UARTDR);
	}
	return n;
}

static int
pl011write(Uart *up, void *buf, int n)
{
	int i;
	char *bp = buf;

	for (i = 0; i < n; i++) {
		while ((plin(up, UARTFR) & FR_TXFF) != 0)
			;
		plout(up, UARTDR, bp[i]);
	}
	return n;
}

static int
pl011signal(Uart *up, int sign, int val)
{
	unsigned cr;

	cr = plin(up, UARTCR);

	switch (sign) {
	case SIG_DTR:
		cr &= ~RTS;
		if (val)
			cr |= RTS;
		up->dtr = (val != 0);
		break;
	case SIG_RTS:
		cr &= ~RTS;
		if (val)
			cr |= RTS;
		up->rts = (val != 0);
		break;
	default:
		return -1;
	}

	plout(up, UARTCR, cr);

	return 1;
}

static int
pl011flush(Uart *up)
{
	unsigned crl;

	crl = plin(up, UARTLCR_H);

	plout(up, UARTLCR_H, crl & ~FEN);
	plout(up, UARTLCR_H, crl);

	return 0;
}

static int
pl011fifo(Uart *up, int trigger)
{
	unsigned fil, cr, val;

        cr = plin(up, UARTCR);

	up->fifo = trigger;
	switch (trigger) {
	case 0:
	case 1:
		plout(up, UARTCR, cr & ~FEN);
		return 0;
	case 4:
		val = 1;
	case 8:
		val = 2;
		break;
	default:
		val = 4;
		break;
	}
	fil = (val << TXIFLSEL) | (val << RXIFLSEL);

	plout(up, UARTCR, cr | FEN);
	plout(up, UARTIFLS, fil);

	return 0;
}

static int
pl011bits(Uart *up, int nbits)
{
	unsigned lcr_h, val;

	lcr_h = plin(up, UARTLCR_H);

	switch (nbits) {
	case 5:
		val = WLEN_5;
		break;
	case 6:
		val = WLEN_6;
		break;
	case 7:
		val = WLEN_7;
		break;
	case 8:
		val = WLEN_8;
		break;
	default:
		return -1;
	}

	up->nbits = val;
	lcr_h &= ~WLEN;
	lcr_h |= val;
	plout(up, UARTLCR_H, lcr_h);

	return 0;
}

static int
pl011parity(Uart *up, int mode)
{
	unsigned lcr_h;

	lcr_h = plin(up, UARTLCR_H);
	lcr_h &= ~(SPS | EPS | PEN);

	switch (mode) {
	case OPARITY:
		lcr_h |= PEN;
		break;
	case EPARITY:
		lcr_h |= PEN | EPS;
		break;
	case NPARITY:
		break;
	default:
		return -1;
	}

	plout(up, UARTLCR_H, lcr_h);

	return 0;
}

static int
pl011stop(Uart *up, int nbits)
{
	unsigned lcr_h;

	lcr_h = plin(up, UARTLCR_H);
	lcr_h &= ~STP2;

	switch (nbits) {
	case 1:
		break;
	case 2:
		lcr_h |= STP2;
		break;
	default:
		return -1;
	}

	up->nbits = nbits;
	plout(up, UARTLCR_H, lcr_h);

	return 0;
}

static int
pl011obey(Uart *up, int signal, int ena)
{
	unsigned cr;

	cr = plin(up, UARTCR);
	ena = (ena != 0);

	switch (signal) {
	case SIG_CTS:
		cr &= ~CTSEN;
		if (ena)
			cr |= CTSEN;
		up->ctsflow = ena;
		break;
	default:
		return -1;
	}

	plout(up, UARTCR, cr);

	return 0;
}

static int
pl011status(Uart *up, struct uartstat *st)
{
	return 0;
}

Uartphy pl011phy = {
	.init = pl011init,
	.setbauds = setbauds,
	.read = pl011read,
	.write = pl011write,
	.signal = pl011signal,
	.flush = pl011flush,
	.setfifo = pl011fifo,
	.numbits = pl011bits,
	.parity = pl011parity,
	.bitstop = pl011stop,
	.obey  = pl011obey,
	.status = pl011status,
};
