#include <os9/os9.h>

#include <libk.h>

#include <ctype.h>
#include <errno.h>
#include <setjmp.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#define PREFIX	"> "
#define NR_ARGC_MAX 5

struct args {
	char *argv[NR_ARGC_MAX];
	int argc;
	int status;
};

struct cmd {
	char *name;
	char *helpmsg;
	int (*eval)(const struct cmd *, struct args *);
	unsigned char min;
	unsigned char max;
};

unsigned in_debug;
jmp_buf dbgrecover;

static const struct cmd cmds[];

static void
error(char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
	kprint(PREFIX "ERR ");
	kvprint(fmt, va);
	kprint("\n");
	longjmp(dbgrecover, 1);
}

static void
ready(void)
{
	kprint(PREFIX "READY\n");
}

/* Get cmd struct for named cmd */
static const struct cmd *
get_cmd(char *name)
{
	const struct cmd *cmd;

	for (cmd = cmds; cmd->name; cmd++) {
		if (strcmp(cmd->name, name) == 0)
			return cmd;
	}
	return NULL;
}

static int
do_exit(const struct cmd *cmd, struct args *args)
{
	args->status = 0;
	if (args->argc == 2)
		args->status = atoi(args->argv[1]);
	return -1;
}

static int
do_help(const struct cmd *cmd, struct args *args)
{
	for (cmd = cmds; cmd->name; cmd++) {
		char *s = cmd->helpmsg;
		kprint("%s\n", s);
	}
	return 0;
}

static int
do_cat(const struct cmd *cmd, struct args *args)
{
	int n;
	Chan *c;
	char buf[100];

	if ((c = devopen(args->argv[1], O_READ)) == NULL)
		goto err;

	while ((n = devread(c, buf, sizeof(buf))) > 0)
		devwrite(console, buf, n);

	if (devclose(c) < 0 || n < 0)
		goto err;

	return 0;

err:
	kprint(PREFIX "ERR cat: %e\n");
	return 0;
}

static int
do_ls(const struct cmd *cmd, struct args *args)
{
	int n;
	Chan *c;
	Dir dentry;
	unsigned char buf[DIRLEN];

	if ((c = devopen(args->argv[1], O_READ)) == NULL)
		goto err;

	while ((n = devread(c, buf, DIRLEN)) > 0) {
		if (p9todir(&dentry, buf, n) < 0)
			goto err;
		kprint("%s\n", dentry.name);
	}

	devclose(c);

	if ( n < 0)
		goto err;

	return 0;

err:
	kprint(PREFIX "ERR ls:%e");
	return 0;
}

static int
do_echo(const struct cmd *cmd, struct args *args)
{
	int i, len, offset, n;
	Chan *c;
	char *s, buffer[LINELEN];

	if ((c = devopen(args->argv[1], O_WRITE)) == NULL)
		goto err;

	offset = 0;
	for (i = 2; i < args->argc; i++) {
		s = args->argv[i];
		len = strlen(s);
		if (len + offset >= LINELEN-2)
			error("line too long");
		memcpy(buffer + offset, s, len);
		offset += len;
		buffer[offset++] = ' ';
	}
	buffer[offset] = '\n';
	n = devwrite(c, buffer, offset);

	if (devclose(c) < 0 || n < 0)
		goto err;

	return 0;
err:
	kprint(PREFIX "ERR echo:%e");
	return 0;
}

static int
do_stat(const struct cmd *cmd, struct args *args)
{
	Dir dir;

	if (dirstat(args->argv[1], &dir) < 0)
		goto err;
	kprint("name: %s\n"
	       "length: %ld\n"
	       "mode: %x\n"
	       "type: %d\n"
	       "dev: %d\n"
	       "qid: %x,%lx,%llx\n",
	       dir.name,
	       dir.length,
	       dir.mode,
	       dir.type,
	       dir.dev,
	       dir.qid.type, dir.qid.vers, dir.qid.path);

	return 0;

err:
	kprint(PREFIX "ERR stat:%e");
	return 0;
}

static int
do_x(const struct cmd *cmd, struct args *args)
{
	char *p, *s;
	int n;
	void *addr;
	size_t len;
	uint16_t *u16;
	uint32_t *u32;
	uint64_t *u64;

	s = args->argv[1];
	if ((p = strchr(s, '/'))  == NULL)
		goto badfmt;
	*p++ = '\0';

	addr = (void *) strtoll(s, NULL, 0);
	n = atoi(p);
	if (n == 0)
		n = 1;

	len = strspn(p, "0123456789");
	switch (p[len]) {
	case 'x':
		u16 = addr;
		while (n-- > 0)
			kprint("%x\n", (int) *u16++);
		return 0;
	case 'X':
		u32 = addr;
		while (n-- > 0)
			kprint("%lx\n", (long) *u32++);
		return 0;
	case 'Y':
		u64 = addr;
		while (n-- > 0)
			kprint("%x\n", (long long) *u64++);
		return 0;
	default:
		goto badfmt;
	}

badfmt:
	kprint(PREFIX "ERR x: bad format");
	return 0;
}

static const struct cmd *
parse_cmd(char *buf, struct args *args)
{
	char *p;
	const struct cmd *cmd;
	const static char ws[] = " \t\r";

	args->argc = 0;
	for (p = strtok(buf, ws); p; p = strtok(NULL, ws)) {
		if (args->argc == NR_ARGC_MAX)
			error("too many parameters");
		args->argv[args->argc++] = p;
	}

	if (args->argc == 0)
		return NULL;

	if ((cmd = get_cmd(args->argv[0])) == NULL)
		error("%s: not found", args->argv[0]);

	return cmd;
}

static int
run(struct args *args)
{
	const struct cmd *cmd;
	size_t len;
	char buffer[LINELEN];

	if (kgets(buffer, sizeof(buffer)) == NULL) {
		args->status = 1;
		return -1;
	}

	if ((len = strlen(buffer)) != 0) {
		if (buffer[len-1] != '\n')
			error("line too long");
		buffer[len-1] = '\0';
		cmd = parse_cmd(buffer, args);
		if (cmd) {
			int (*fn)(const struct cmd *, struct args *);
			void *bp;

			if (args->argc < cmd->min || args->argc > cmd->max)
				error(cmd->helpmsg);
			bp = cmd->eval;
			fn = (int (*)(const struct cmd *, struct args *)) bp;
			return (*fn)(cmd, args);
		}
	}

	return 0;
}

int
debug(void)
{
	struct args args;

	kprint("begin debug language interface\n");
	in_debug = 1;
	setjmp(dbgrecover);

	for (ready(); !run(&args); ready())
		;
	kprint("end debug language interface\n");
	in_debug = 0;

	return args.status;
}

/*
 * definition of arrays
 */
static const struct cmd cmds[] = {
	{
		.name = "exit",
		.eval = do_exit,
		.min = 1,
		.max = 2,
		.helpmsg = "Exit debug environment: exit [status]",
	},
	{
		.name = "help",
		.eval = do_help,
		.min = 1,
		.max = 1,
		.helpmsg = "Print this help menu: help",
	},
	{
		.name = "ls",
		.eval = do_ls,
		.min = 2,
		.max = 2,
		.helpmsg = "List content of a directory: ls path",
	},
	{
		.name = "cat",
		.eval = do_cat,
		.min = 2,
		.max = 2,
		.helpmsg = "Read content from a file: cat path",
	},
	{
		.name = "echo",
		.eval = do_echo,
		.min = 3,
		.max = NR_ARGC_MAX,
		.helpmsg = "Write content to a file: echo path string ...",
	},
	{
		.name = "stat",
		.eval = do_stat,
		.min = 2,
		.max = 2,
		.helpmsg = "Display stat attribute: stat path",
	},
	{
		.name = "x",
		.eval = do_x,
		.min = 2,
		.max = 2,
		.helpmsg = "Examine memory: x addr/fmt",
	},
	{
		.name = NULL
	}
};
