#include <os9/os9.h>

#include <errno.h>
#include <limits.h>

struct nspace {
	Ref ref;
	mutex_t m;
	struct mpoint mpoints[NR_MPOINTS];
};

struct fdset {
	Ref ref;
	mutex_t m;
	Chan fdset[NR_CHANS];
};

static Task tasktab[NR_TASKS];

static void
initref(Ref *rp)
{
	rp->m = 0;
	rp->cnt = 0;
}

static void
incref(Ref *rp)
{
	lock(&rp->m);
	rp->cnt++;
	unlock(&rp->m);
}

static void
decref(Ref *rp)
{
	lock(&rp->m);
	rp->cnt--;
	unlock(&rp->m);
}

int
getntask(int n, Task **tpp)
{
	Task *tp;

	if (n >= NR_TASKS)
		return -1;
	tp = &tasktab[n];

	if (tp->mode == PDISABLED)
		return 0;

	*tpp	= tp;
	return 1;
}

Task *
gettask(int tid)
{
	Task *tp;

	for (tp = tasktab; tp < &tasktab[NR_TASKS]; ++tp) {
		if (tp->mode != PDISABLED && tp->tid == tid)
			break;
	}

	return (tp != &tasktab[NR_TASKS]) ? tp : NULL;
}

static int
newtid(void)
{
	int tid, id;
	static int last;

	id = tid = last;

	do {
		if (tid == INT_MAX)
			tid = -1;
		tid++;
		if (gettask(tid) == NULL)
			break;
	} while (tid != last);

	last = tid;

	return (tid != id) ? tid : -1;
}

static Task *
newslot(void)
{
	Task *tp;

	for (tp = tasktab; tp < &tasktab[NR_TASKS]; ++tp) {
		if (tp->mode == PDISABLED)
			break;
	}

	return (tp != &tasktab[NR_TASKS]) ? tp : NULL;
}

static Nspace *
newspace(Nspace *from)
{
	Nspace *ns;

	ns = alloc(sizeof(*ns));
	if (from)
		*ns = *from;
	initref(&ns->ref);

	return ns;
}

static Fdset *
newfds(Fdset *from)
{
	Fdset *fds;

	fds = alloc(sizeof(*fds));
	if (from)
		*fds = *from;
	initref(&fds->ref);

	return fds;
}

static Map *
newmap(Map *from)
{
	Map *mp;

	mp = alloc(sizeof(*mp));
	if (from)
		*mp = *from;
	initref(&mp->ref);

	return mp;
}

int
rfork(int flags)
{
	Task *tp, *parent;

	tp = parent = task;

	if ((flags & (RFPROC|RFMEM))  == (RFPROC|RFMEM)
	||   (flags & (RFNAMEG|RFCNAMEG))  == (RFNAMEG|RFCNAMEG)
	||   (flags & (RFFDG|RFCFDG))  == (RFFDG|RFCFDG)) {
		errno = EINVAL;
		return -1;
	}

	if ((flags & RFPROC) != 0) {
		if ((tp = newslot()) == NULL)
			goto enomem;
		if ((tp->tid = newtid()) < 0)
			goto enomem;

		incref(&parent->text->ref);
		tp->text = parent->text;

		incref(&parent->data->ref);
		tp->data = parent->data;

		incref(&parent->stack->ref);
		tp->stack = parent->stack;

		incref(&parent->ns->ref);
		tp->ns = parent->ns;

		incref(&parent->fds->ref);
		tp->fds = parent->fds;
	}

	if ((flags & RFCNAMEG) != 0) {
		decref(&tp->ns->ref);
		tp->ns = newspace(NULL);
		incref(&tp->ns->ref);
	}
	if ((flags & RFNAMEG) != 0) {
		decref(&tp->ns->ref);
		tp->ns = newspace(parent->ns);
		incref(&tp->ns->ref);
	}
	if ((flags & RFCFDG) != 0) {
		decref(&tp->fds->ref);
		tp->fds = newfds(NULL);
		incref(&tp->fds->ref);
	}
	if ((flags & RFFDG) != 0) {
		decref(&tp->fds->ref);
		tp->fds = newfds(parent->fds);
		incref(&tp->fds->ref);
	}
	if ((flags & RFCFDG) != 0) {
		decref(&tp->data->ref);
		tp->data = newmap(parent->data);
		incref(&tp->data->ref);
	}

	return (flags & RFPROC) ? tp->tid : 0;

enomem:
	errno = ENOMEM;
	return -1;
}
